import React, { useEffect, useRef, useState } from 'react';
import { useFrame } from '@react-three/fiber';
import * as THREE from 'three';
import {
    useAnimationPlaySpeed,
    usePauseButtonStore,
    usePlayButtonStore,
    useResetButtonStore
} from '../../../../../store/usePlayButtonStore';
import { Line } from '@react-three/drei';

function RoboticArmAnimator({
    HandleCallback,
    restPosition,
    ikSolver,
    targetBone,
    armBot,
    logStatus,
    path
}: any) {
    const progressRef = useRef(0);
    const curveRef = useRef<THREE.Vector3[] | null>(null);
    const [currentPath, setCurrentPath] = useState<[number, number, number][]>([]);
    const [curvePoints, setCurvePoints] = useState<THREE.Vector3[] | null>(null);

    // Zustand stores
    const { isPlaying } = usePlayButtonStore();
    const { isPaused } = usePauseButtonStore();
    const { isReset } = useResetButtonStore(); // optional use depending on your logic
    const { speed } = useAnimationPlaySpeed();

    useEffect(() => {
        setCurrentPath(path)
    }, [path])

    useEffect(() => {
        if (!isPlaying) {
            setCurrentPath([])
            curveRef.current = null
        }
    }, [isPlaying])

    useEffect(() => {
        if (currentPath && currentPath.length === 3) {
            const [start, mid, end] = currentPath;
            const points = [
                new THREE.Vector3(start[0], start[1], start[2]),
                new THREE.Vector3(start[0], mid[1] + 0.5, start[2]),
                new THREE.Vector3(mid[0], mid[1] + 0.5, mid[2]),
                new THREE.Vector3(mid[0], end[1] + 0.5, mid[2]),
                new THREE.Vector3(end[0], end[1], end[2])
            ];
            const generatedCurve = new THREE.CatmullRomCurve3(points, false, 'centripetal', 1).getPoints(100);
            // console.log('generatedCurve: ', generatedCurve);
            curveRef.current = generatedCurve;
            setCurvePoints(generatedCurve);

        }
    }, [currentPath]);

    useFrame((_, delta) => {
        if (ikSolver && curveRef.current && currentPath.length >= 0) {
            const bone = ikSolver.mesh.skeleton.bones.find(
                (b: any) => b.name === targetBone
            );
            if (!bone) return;

            if (isPlaying) {
                if (!isPaused) {
                    const curvePoints = curveRef.current;
                    const speedAdjustedProgress = progressRef.current + (speed * armBot.speed);
                    const index = Math.floor(speedAdjustedProgress);
                    if (index >= curvePoints.length) {
                        HandleCallback();
                        setCurrentPath([]);
                        curveRef.current = null;
                        progressRef.current = 0;
                    } else {
                        const point = curvePoints[index];
                        bone.position.copy(point);
                        progressRef.current = speedAdjustedProgress;
                    }
                }
            } else {
                logStatus(armBot.modelUuid, 'Simulation Play Exited');
                // bone.position.copy(restPosition);
            }

            ikSolver.update();
        } else if (ikSolver && !isPlaying && currentPath.length === 0) {
            const bone = ikSolver.mesh.skeleton.bones.find(
                (b: any) => b.name === targetBone
            );
            // bone.position.copy(restPosition);
            ikSolver.update();
        };

    });

    return (
        <>
            {/* {currentPath.length > 0 && (
                <mesh rotation={armBot.rotation}>
                    <Line points={currentPath} color="green" lineWidth={3} />
                    {currentPath.map((point, index) => (
                        <mesh key={index} position={point} rotation={armBot.rotation}>
                            <sphereGeometry args={[0.1, 16, 16]} />
                            <meshStandardMaterial color="red" />
                        </mesh>
                    ))}
                </mesh>
            )} */}
            {curvePoints && (
                <mesh rotation={armBot.rotation}>
                    <Line
                        points={curvePoints.map(p => [p.x, p.y, p.z])}
                        color="green"
                        lineWidth={5}
                        dashed={false}
                    />
                    {currentPath.length >= 1 && (
                        <>
                            {/* First point */}
                            <mesh position={currentPath[0]} rotation={armBot.rotation}>
                                <sphereGeometry args={[0.1, 16, 16]} />
                                <meshStandardMaterial color="red" />
                            </mesh>

                            {/* Last point */}
                            <mesh position={currentPath[currentPath.length - 1]} rotation={armBot.rotation}>
                                <sphereGeometry args={[0.1, 16, 16]} />
                                <meshStandardMaterial color="red" />
                            </mesh>
                        </>
                    )}
                </mesh>
            )}
        </>
    );
}

export default RoboticArmAnimator;
